@@ -39,15 +39,23 @@
 #![no_std]
 #![feature(generic_associated_types)]
 #![feature(iter_order_by)]
-
-use core::convert::TryInto;
+#![cfg_attr(
+    feature = "nontrivial_option_processing",
+    feature(type_alias_impl_trait)
+)]
 
 use coap_message::{MessageOption, MutableWritableMessage, ReadableMessage};
 
 pub mod helpers;
+pub mod option_processing;
+pub mod wkc;
+mod wkc_implementation;
+
+mod forking_helpers;
+use forking_helpers::{ForkingRecord, PrefixedRecord};
 
 use crate::helpers::{
-    block2_write_with_cf, codeconvert, Block2RequestData, MaskingUriUpToPath,
+    block2_write_with_cf, codeconvert, Block2RequestData, MaskingUriUpToPath, MaskingUriUpToPathN,
 };
 use coap_handler::Handler;
 
@@ -91,17 +99,18 @@ impl Handler for NeverFound {
 /// Start building a tree of sub-resources
 ///
 /// While technically this just returns a handler that returns 4.04 unconditionally, it also
-/// implemnts HandlerBuilder, and thus can be used like this:
+/// implements HandlerBuilder, and thus can be used like this:
 ///
 /// ```
 /// use coap_handler_implementations::*;
 /// let handler = new_dispatcher()
 ///     .at(&["dev", "name"], SimpleRendered::new_typed_str("Demo program", Some(0)))
 ///     .at(&["dev", "version"], SimpleRendered::new_typed_str("0.8.15", Some(0)))
+///     .with_wkc()
 ///     ;
 /// ```
-pub fn new_dispatcher() -> NeverFound {
-    NeverFound {}
+pub fn new_dispatcher() -> impl Handler + wkc::Reporting {
+    wkc::NotReporting::new(NeverFound {})
 }
 
 /// Trait implemented by default on all handlers that lets the user stack them using a builder-like
@@ -120,9 +129,84 @@ where
     /// The handler will not *not* see the Uri-Path (and Uri-Host, as this builder doesn't do
     /// virtual hosting yet) options any more; see the top-level module documentation on Options
     /// Hiding for rationale.
+    ///
+    /// If both the previous tree and the new handler are Reporting, so is the result.
     fn at<H>(self, path: &'a [&'a str], handler: H) -> ForkingHandler<'a, H, Self>
     where
-        H: Handler + Sized;
+        H: Handler + Sized,
+    {
+        ForkingHandler {
+            h1: handler,
+            h2: self,
+            h1_condition: path,
+        }
+    }
+
+    /// Divert requests arriving at `path` into the given `handler`, and announce them with the
+    /// given attributes in .well-known/core.
+    ///
+    /// Any reporting the handler would have done is overridden.
+    ///
+    /// This is a shorthand for `.at(ConstantSingleRecordReport::new(h, attributes))`.
+    fn at_with_attributes<H>(
+        self,
+        path: &'a [&'a str],
+        attributes: &'a [wkc::Attribute],
+        handler: H,
+    ) -> ForkingHandler<'a, wkc::ConstantSingleRecordReport<'a, H>, Self>
+    where
+        H: Handler + Sized,
+    {
+        ForkingHandler {
+            h1: wkc::ConstantSingleRecordReport::new(handler, attributes),
+            h2: self,
+            h1_condition: path,
+        }
+    }
+
+    /// Divert requests arriving with an Uri-Path starting with `path` to the given `handler`.
+    ///
+    /// Only remaining Uri-Path options will be visible to the handler; those expressed in path
+    /// (and Uri-Host, see [.at()]) are hidden.
+    ///
+    /// If both the previous tree and the new handler are Reporting, so is the result.
+    fn below<H>(self, path: &'a [&'a str], handler: H) -> ForkingTreeHandler<'a, H, Self> {
+        ForkingTreeHandler {
+            h1: handler,
+            h2: self,
+            h1_condition: path,
+        }
+    }
+}
+
+impl<'a, OldRD, OldH> HandlerBuilder<'a, OldRD> for OldH
+where
+    Self: Handler<RequestData = OldRD> + Sized,
+{
+    // Methods are provided
+}
+
+/// Extension trait for handlers that also implement [Reporting](crate::wkc::Reporting).
+///
+/// Like [HandlerBuilder] this is implemented for wherever it works.
+///
+/// (Note that while this *could* be implemented as a provided method of Reporting, it is split out
+/// to stay architecturally analogous to the HandlerBuilder, and to not force this crate's
+/// implementation of .well-known/core onto other users of Reporting. Possibly, these could be
+/// separated approaching stabilization.)
+pub trait ReportingHandlerBuilder<'a, OldRD>: HandlerBuilder<'a, OldRD> + wkc::Reporting {
+    /// Add a `/.well-known/core` resource that exposes the information the previous (stack of)
+    /// handler(s) exposes throught the [Reporting](crate::wkc::Reporting) trait.
+    fn with_wkc(self) -> wkc_implementation::WellKnownCore<Self> {
+        wkc_implementation::WellKnownCore::new(self)
+    }
+}
+
+impl<'a, OldRD, OldH> ReportingHandlerBuilder<'a, OldRD> for OldH
+where
+    OldH: Handler<RequestData = OldRD> + wkc::Reporting,
+{
+    // Methods are provided
 }
 
 pub struct ForkingHandler<'a, H1, H2> {
@@ -180,68 +264,87 @@ where
     }
 }
 
-impl<'a, OldRD, OldH> HandlerBuilder<'a, OldRD> for OldH
+impl<'a, RD1, H1, RD2, H2> wkc::Reporting for ForkingHandler<'a, H1, H2>
 where
-    Self: Handler<RequestData = OldRD> + Sized,
+    H1: Handler<RequestData = RD1> + wkc::Reporting,
+    H2: Handler<RequestData = RD2> + wkc::Reporting,
 {
-    fn at<H>(self, path: &'a [&'a str], handler: H) -> ForkingHandler<'a, H, Self> {
-        ForkingHandler {
-            h1: handler,
-            h2: self,
-            h1_condition: path,
+    // FIXME: This is copied over from ForkingTreeHandler (and stripped to the general not(feature
+    // = "nontrivial_option_processing") case)
+    //
+    // As it is now it would appear to warrant the Forking{,Tree}Handler unification hinted at the
+    // ForkingTreeHandler definition. Not starting the deduplication (which would be warranted by
+    // the below behemoth) as this'll all look much slimmer again once type_alias_impl_trait is
+    // on by default.
+    //
+    // (After that, could be that it's warranted as it's now more than a few 4-liners, could be
+    // that not).
+
+    type Record = ForkingRecord<PrefixedRecord<'a, H1::Record>, H2::Record>;
+
+    type Reporter = core::iter::Chain<
+        core::iter::Map<H2::Reporter, fn(H2::Record) -> Self::Record>,
+        core::iter::Map<
+            core::iter::Zip<H1::Reporter, core::iter::Repeat<&'a [&'a str]>>,
+            fn((H1::Record, &'a [&'a str])) -> Self::Record,
+        >,
+    >;
+    fn report(&self) -> Self::Reporter {
+        fn first<'c, H1R, H2R>(
+            prefixed_prefix: (H1R, &'c [&'c str]),
+        ) -> ForkingRecord<PrefixedRecord<'c, H1R>, H2R> {
+            let (prefixed, prefix) = prefixed_prefix;
+            ForkingRecord::First(PrefixedRecord { prefix, prefixed })
         }
+        self.h2
+            .report()
+            .map(ForkingRecord::Second as fn(_) -> _)
+            .chain(
+                self.h1
+                    .report()
+                    .zip(core::iter::repeat(self.h1_condition))
+                    .map(first as fn(_) -> _),
+            )
     }
 }
 
-#[deprecated(note="Use SimpleRendered::new_typed_str/_slice")]
-pub struct TypedStaticResponse {
-    pub payload: &'static [u8],
-    pub contentformat: &'static [u8], // pre-encoded
+// This is identical to the ForkingHandler in its structure -- just the matching behavior differs.
+// Even ForkingRequestData can be shared; unfortunately, the main code is still duplicated -- it
+// could be refactored, but is it worth it for the identical 4-lines parts?
+pub struct ForkingTreeHandler<'a, H1, H2> {
+    h1: H1,
+    h2: H2,
+
+    h1_condition: &'a [&'a str],
 }
 
-#[allow(deprecated)] // It's deprecated for users
-impl From<&'static str> for TypedStaticResponse {
-    fn from(s: &'static str) -> Self {
-        TypedStaticResponse {
-            payload: s.as_bytes(),
-            contentformat: &[], // 0; text/plain;charset=utf-8
-        }
-    }
-}
-
-#[allow(deprecated)] // It's deprecated for users
-impl Handler for TypedStaticResponse {
-    type RequestData = Code;
+impl<'a, RD1, H1, RD2, H2> Handler for ForkingTreeHandler<'a, H1, H2>
+where
+    H1: Handler<RequestData = RD1>,
+    H2: Handler<RequestData = RD2>,
+{
+    type RequestData = ForkingRequestData<RD1, RD2>;
 
     fn extract_request_data(&mut self, request: &impl ReadableMessage) -> Self::RequestData {
-        match request.code().into() {
-            code::GET => (),
-            _ => return code::METHOD_NOT_ALLOWED,
-        };
+        let expected_path = self.h1_condition.iter().map(|s| s.as_bytes());
+        let actual_path = request
+            .options()
+            .filter(|o| o.number() == option::URI_PATH)
+            .take(self.h1_condition.len());
 
-        for o in request.options() {
-            match o.number() {
-                option::ACCEPT => {
-                    if o.value() != self.contentformat {
-                        return code::UNSUPPORTED_CONTENT_FORMAT;
-                    }
-                }
-                o if option::get_criticality(o) == option::Criticality::Critical => {
-                    return code::BAD_OPTION;
-                }
-                _ => {}
-            }
+        if expected_path.cmp_by(actual_path, |e, a| e.cmp(a.value())) == core::cmp::Ordering::Equal
+        {
+            let masked = MaskingUriUpToPathN::new(request, self.h1_condition.len());
+            ForkingRequestData::First(self.h1.extract_request_data(&masked))
+        } else {
+            ForkingRequestData::Second(self.h2.extract_request_data(request))
         }
-
-        code::CONTENT
     }
 
     fn estimate_length(&mut self, request: &Self::RequestData) -> usize {
-        match *request {
-            code::CONTENT => {
-                2 /* small option offset plus payload marker */ + self.payload.len() + self.contentformat.len()
-            }
-            _ => 0,
+        match request {
+            ForkingRequestData::First(r) => self.h1.estimate_length(r),
+            ForkingRequestData::Second(r) => self.h2.estimate_length(r),
         }
     }
 
@@ -250,15 +353,62 @@ impl Handler for TypedStaticResponse {
         response: &mut impl MutableWritableMessage,
         request: Self::RequestData,
     ) {
-        response.set_code(codeconvert(request));
-        let mut p = &b""[..];
-        if request == code::CONTENT {
-            if let Ok(cfopt) = option::CONTENT_FORMAT.try_into() {
-                response.add_option(cfopt, self.contentformat);
-            }
-            p = self.payload;
+        match request {
+            ForkingRequestData::First(r) => self.h1.build_response(response, r),
+            ForkingRequestData::Second(r) => self.h2.build_response(response, r),
         }
-        response.set_payload(p);
+    }
+}
+
+impl<'a, RD1, H1, RD2, H2> wkc::Reporting for ForkingTreeHandler<'a, H1, H2>
+where
+    H1: Handler<RequestData = RD1> + wkc::Reporting,
+    H2: Handler<RequestData = RD2> + wkc::Reporting,
+{
+    type Record = ForkingRecord<PrefixedRecord<'a, H1::Record>, H2::Record>;
+
+    // FIXME: one of these is copied over to wkc::Reporting for ForkingHandler, see there.
+
+    #[cfg(feature = "nontrivial_option_processing")]
+    type Reporter = impl Iterator<Item = Self::Record>;
+    #[cfg(feature = "nontrivial_option_processing")]
+    fn report(&self) -> Self::Reporter {
+        self.h2
+            .report()
+            .map(ForkingRecord::Second as fn(_) -> _)
+            .chain(self.h1.report().map(|f| {
+                ForkingRecord::First(PrefixedRecord {
+                    prefix: self.h1_condition,
+                    prefixed: f,
+                })
+            }))
+    }
+
+    #[cfg(not(feature = "nontrivial_option_processing"))]
+    type Reporter = core::iter::Chain<
+        core::iter::Map<H2::Reporter, fn(H2::Record) -> Self::Record>,
+        core::iter::Map<
+            core::iter::Zip<H1::Reporter, core::iter::Repeat<&'a [&'a str]>>,
+            fn((H1::Record, &'a [&'a str])) -> Self::Record,
+        >,
+    >;
+    #[cfg(not(feature = "nontrivial_option_processing"))]
+    fn report(&self) -> Self::Reporter {
+        fn first<'c, H1R, H2R>(
+            prefixed_prefix: (H1R, &'c [&'c str]),
+        ) -> ForkingRecord<PrefixedRecord<'c, H1R>, H2R> {
+            let (prefixed, prefix) = prefixed_prefix;
+            ForkingRecord::First(PrefixedRecord { prefix, prefixed })
+        }
+        self.h2
+            .report()
+            .map(ForkingRecord::Second as fn(_) -> _)
+            .chain(
+                self.h1
+                    .report()
+                    .zip(core::iter::repeat(self.h1_condition))
+                    .map(first as fn(_) -> _),
+            )
     }
 }
 
@@ -280,8 +430,7 @@ pub trait SimpleRenderable {
     where
         W: core::fmt::Write;
 
-    fn render_bytes(&mut self, writer: &mut helpers::WindowedInfinityWithETag)
-    {
+    fn render_bytes(&mut self, writer: &mut helpers::WindowedInfinityWithETag) {
         self.render(writer)
     }
 
@@ -297,7 +446,10 @@ pub struct SimpleRendered<T: SimpleRenderable>(pub T);
 
 impl<'a> SimpleRendered<TypedStaticRenderable<'a>> {
     pub fn new_typed_slice(data: &'a [u8], content_format: Option<u16>) -> Self {
-        SimpleRendered(TypedStaticRenderable { data, content_format })
+        SimpleRendered(TypedStaticRenderable {
+            data,
+            content_format,
+        })
     }
 
     pub fn new_typed_str(data: &'a str, content_format: Option<u16>) -> Self {
@@ -330,7 +482,7 @@ where
                         Ok(Some(_)) => Err(coap_numbers::code::BAD_REQUEST),
                         Ok(None) => Block2RequestData::from_option(&o)
                             .map(Some)
-                            .map_err(|()| code::BAD_REQUEST),
+                            .map_err(|_| code::BAD_REQUEST),
                     }
                 }
                 o if option::get_criticality(o) == option::Criticality::Critical => {
@@ -371,6 +523,20 @@ where
     }
 }
 
+impl<T> wkc::Reporting for SimpleRendered<T>
+where
+    T: SimpleRenderable,
+{
+    type Record = wkc::EmptyRecord;
+    type Reporter = core::iter::Once<wkc::EmptyRecord>;
+
+    fn report(&self) -> Self::Reporter {
+        // Using a ConstantSliceRecord instead would be tempting, but that'd need a const return
+        // value from self.0.content_format()
+        core::iter::once(wkc::EmptyRecord {})
+    }
+}
+
 impl<'a> SimpleRenderable for &'a str {
     fn render<W>(&mut self, writer: &mut W)
     where
@@ -388,7 +554,7 @@ impl<'a> SimpleRenderable for &'a str {
 
 pub struct TypedStaticRenderable<'a> {
     data: &'a [u8],
-    content_format: Option<u16>
+    content_format: Option<u16>,
 }
 
 impl<'a> SimpleRenderable for TypedStaticRenderable<'a> {
@@ -470,7 +636,7 @@ impl<H: SimpleCBORHandler> SimpleCBORWrapper<H> {
                         Ok(Some(_)) => Err(coap_numbers::code::BAD_REQUEST),
                         Ok(None) => Block2RequestData::from_option(&o)
                             .map(Some)
-                            .map_err(|()| code::BAD_REQUEST),
+                            .map_err(|_| code::BAD_REQUEST),
                     }
                 }
                 o if option::get_criticality(o) == option::Criticality::Critical => {
@@ -518,12 +684,10 @@ where
 
     fn extract_request_data(&mut self, request: &impl ReadableMessage) -> Self::RequestData {
         match request.code().into() {
-            code::GET => {
-                match Self::check_get_options(request) {
-                    Err(e) => Done(e),
-                    Ok(block) => Get(block),
-                }
-            }
+            code::GET => match Self::check_get_options(request) {
+                Err(e) => Done(e),
+                Ok(block) => Get(block),
+            },
             code::POST => {
                 if let Err(e) = Self::check_postput_options(request) {
                     return Done(e);
